@@ -235,7 +235,7 @@ g_tls_connection_gnutls_initable_init (GInitable     *initable,
   gnutls_transport_set_ptr (priv->session, gnutls);
 
   /* GDatagramBased supports vectored I/O; GPollableOutputStream does not. */
-  if (base_socket != NULL)
+  if (base_socket)
     {
       gnutls_transport_set_vec_push_function (priv->session,
                                               g_tls_connection_gnutls_vec_push_func);
@@ -347,14 +347,21 @@ end_gnutls_io (GTlsConnectionGnutls  *gnutls,
   if (handshaking && !ever_handshaked)
     {
       if (g_error_matches (my_error, G_IO_ERROR, G_IO_ERROR_FAILED) ||
-          g_error_matches (my_error, G_IO_ERROR, G_IO_ERROR_BROKEN_PIPE) ||
-          status == GNUTLS_E_UNEXPECTED_PACKET_LENGTH ||
+          g_error_matches (my_error, G_IO_ERROR, G_IO_ERROR_BROKEN_PIPE))
+        {
+          g_set_error (error, G_TLS_ERROR, G_TLS_ERROR_NOT_TLS,
+                       _("Peer failed to perform TLS handshake: %s"), my_error->message);
+          g_clear_error (&my_error);
+          return G_TLS_CONNECTION_BASE_ERROR;
+        }
+
+      if (status == GNUTLS_E_UNEXPECTED_PACKET_LENGTH ||
           status == GNUTLS_E_DECRYPTION_FAILED ||
           status == GNUTLS_E_UNSUPPORTED_VERSION_PACKET)
         {
           g_clear_error (&my_error);
-          g_set_error_literal (error, G_TLS_ERROR, G_TLS_ERROR_NOT_TLS,
-                               _("Peer failed to perform TLS handshake"));
+          g_set_error (error, G_TLS_ERROR, G_TLS_ERROR_NOT_TLS,
+                       _("Peer failed to perform TLS handshake: %s"), gnutls_strerror (status));
           return G_TLS_CONNECTION_BASE_ERROR;
         }
     }
@@ -377,8 +384,8 @@ end_gnutls_io (GTlsConnectionGnutls  *gnutls,
       if (handshaking && !ever_handshaked)
         {
           g_clear_error (&my_error);
-          g_set_error_literal (error, G_TLS_ERROR, G_TLS_ERROR_NOT_TLS,
-                               _("Peer failed to perform TLS handshake"));
+          g_set_error (error, G_TLS_ERROR, G_TLS_ERROR_NOT_TLS,
+                       _("Peer failed to perform TLS handshake: %s"), gnutls_strerror (status));
           return G_TLS_CONNECTION_BASE_ERROR;
         }
 
@@ -408,8 +415,8 @@ end_gnutls_io (GTlsConnectionGnutls  *gnutls,
   if (ret == GNUTLS_E_CERTIFICATE_ERROR)
     {
       g_clear_error (&my_error);
-      g_set_error (error, G_TLS_ERROR, G_TLS_ERROR_BAD_CERTIFICATE,
-                   _("Unacceptable TLS certificate"));
+      g_set_error_literal (error, G_TLS_ERROR, G_TLS_ERROR_BAD_CERTIFICATE,
+                           _("Unacceptable TLS certificate"));
       return G_TLS_CONNECTION_BASE_ERROR;
     }
 
@@ -750,11 +757,21 @@ g_tls_connection_gnutls_pull_timeout_func (gnutls_transport_ptr_t transport_data
   return 0;
 }
 
+static GTlsSafeRenegotiationStatus
+g_tls_connection_gnutls_handshake_thread_safe_renegotiation_status (GTlsConnectionBase *tls)
+{
+  GTlsConnectionGnutls *gnutls = G_TLS_CONNECTION_GNUTLS (tls);
+  GTlsConnectionGnutlsPrivate *priv = g_tls_connection_gnutls_get_instance_private (gnutls);
+
+  return gnutls_safe_renegotiation_status (priv->session) ? G_TLS_SAFE_RENEGOTIATION_SUPPORTED_BY_PEER
+                                                          : G_TLS_SAFE_RENEGOTIATION_UNSUPPORTED;
+}
+
 static GTlsConnectionBaseStatus
-g_tls_connection_gnutls_request_rehandshake (GTlsConnectionBase  *tls,
-                                             gint64               timeout,
-                                             GCancellable        *cancellable,
-                                             GError             **error)
+g_tls_connection_gnutls_handshake_thread_request_rehandshake (GTlsConnectionBase  *tls,
+                                                              gint64               timeout,
+                                                              GCancellable        *cancellable,
+                                                              GError             **error)
 {
   GTlsConnectionGnutls *gnutls = G_TLS_CONNECTION_GNUTLS (tls);
   GTlsConnectionGnutlsPrivate *priv = g_tls_connection_gnutls_get_instance_private (gnutls);
@@ -868,7 +885,7 @@ g_tls_connection_gnutls_handshake_thread_handshake (GTlsConnectionBase  *tls,
       ret = gnutls_record_recv (priv->session, buf, sizeof (buf));
       if (ret > -1)
         {
-          g_tls_connection_base_buffer_application_data (tls, buf, ret);
+          g_tls_connection_base_handshake_thread_buffer_application_data (tls, buf, ret);
           ret = GNUTLS_E_AGAIN;
         }
     }
@@ -921,8 +938,7 @@ g_tls_connection_gnutls_read (GTlsConnectionBase  *tls,
   ret = gnutls_record_recv (priv->session, buffer, count);
   END_GNUTLS_IO (gnutls, G_IO_IN, ret, status, _("Error reading data from TLS socket"), error);
 
-  if (ret >= 0)
-    *nread = ret;
+  *nread = MAX (ret, 0);
   return status;
 }
 
@@ -982,8 +998,7 @@ g_tls_connection_gnutls_read_message (GTlsConnectionBase  *tls,
 
   END_GNUTLS_IO (gnutls, G_IO_IN, ret, status, _("Error reading data from TLS socket"), error);
 
-  if (ret >= 0)
-    *nread = ret;
+  *nread = MAX (ret, 0);
   return status;
 }
 
@@ -1005,8 +1020,7 @@ g_tls_connection_gnutls_write (GTlsConnectionBase  *tls,
   ret = gnutls_record_send (priv->session, buffer, count);
   END_GNUTLS_IO (gnutls, G_IO_OUT, ret, status, _("Error writing data to TLS socket"), error);
 
-  if (ret >= 0)
-    *nwrote = ret;
+  *nwrote = MAX (ret, 0);
   return status;
 }
 
@@ -1070,8 +1084,7 @@ g_tls_connection_gnutls_write_message (GTlsConnectionBase  *tls,
   ret = gnutls_record_uncork (priv->session, 0  /* flags */);
   END_GNUTLS_IO (gnutls, G_IO_OUT, ret, status, _("Error writing data to TLS socket"), error);
 
-  if (ret > 0)
-    *nwrote = ret;
+  *nwrote = MAX (ret, 0);
   return status;
 }
 
@@ -1099,19 +1112,20 @@ g_tls_connection_gnutls_class_init (GTlsConnectionGnutlsClass *klass)
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   GTlsConnectionBaseClass *base_class = G_TLS_CONNECTION_BASE_CLASS (klass);
 
-  gobject_class->finalize                = g_tls_connection_gnutls_finalize;
+  gobject_class->finalize                                = g_tls_connection_gnutls_finalize;
 
-  base_class->request_rehandshake        = g_tls_connection_gnutls_request_rehandshake;
-  base_class->prepare_handshake          = g_tls_connection_gnutls_prepare_handshake;
-  base_class->handshake_thread_handshake = g_tls_connection_gnutls_handshake_thread_handshake;
-  base_class->retrieve_peer_certificate  = g_tls_connection_gnutls_retrieve_peer_certificate;
-  base_class->complete_handshake         = g_tls_connection_gnutls_complete_handshake;
-  base_class->is_session_resumed         = g_tls_connection_gnutls_is_session_resumed;
-  base_class->read_fn                    = g_tls_connection_gnutls_read;
-  base_class->read_message_fn            = g_tls_connection_gnutls_read_message;
-  base_class->write_fn                   = g_tls_connection_gnutls_write;
-  base_class->write_message_fn           = g_tls_connection_gnutls_write_message;
-  base_class->close_fn                   = g_tls_connection_gnutls_close;
+  base_class->prepare_handshake                          = g_tls_connection_gnutls_prepare_handshake;
+  base_class->handshake_thread_safe_renegotiation_status = g_tls_connection_gnutls_handshake_thread_safe_renegotiation_status;
+  base_class->handshake_thread_request_rehandshake       = g_tls_connection_gnutls_handshake_thread_request_rehandshake;
+  base_class->handshake_thread_handshake                 = g_tls_connection_gnutls_handshake_thread_handshake;
+  base_class->retrieve_peer_certificate                  = g_tls_connection_gnutls_retrieve_peer_certificate;
+  base_class->complete_handshake                         = g_tls_connection_gnutls_complete_handshake;
+  base_class->is_session_resumed                         = g_tls_connection_gnutls_is_session_resumed;
+  base_class->read_fn                                    = g_tls_connection_gnutls_read;
+  base_class->read_message_fn                            = g_tls_connection_gnutls_read_message;
+  base_class->write_fn                                   = g_tls_connection_gnutls_write;
+  base_class->write_message_fn                           = g_tls_connection_gnutls_write_message;
+  base_class->close_fn                                   = g_tls_connection_gnutls_close;
 }
 
 static void
